#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__init__.py

@author : Louis RICHARD
"""

from .ts_time import ts_time
from .ts_scalar import ts_scalar
from .ts_vec_xyz import ts_vec_xyz
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_skymap import ts_skymap
from .ts_append import ts_append
from .dist_append import dist_append
from .start import start
from .end import end
from .iso2unix import iso2unix
from .extend_tint import extend_tint
from .dot import dot
from .cross import cross
from .trace import trace
from .norm import norm
from .normalize import normalize
from .calc_dt import calc_dt
from .calc_fs import calc_fs
from .avg_4sc import avg_4sc
from .gradient import gradient
from .integrate import integrate
from .time_clip import time_clip
from .resample import resample
from .t_eval import t_eval
from .filt import filt
from .medfilt import medfilt
from .movmean import movmean
from .wavelet import wavelet
from .mva import mva
from .vht import vht
from .new_xyz import new_xyz
from .dec_par_perp import dec_par_perp
from .convert_fac import convert_fac
from .edb import edb
from .e_vxb import e_vxb
from .plasma_calc import plasma_calc
from .dynamic_press import dynamic_press
from .pres_anis import pres_anis
from .agyro_coeff import agyro_coeff
from .median_bins import median_bins
from .mean_bins import mean_bins
from .histogram import histogram
from .histogram2d import histogram2d
from .calc_disprel_tm import calc_disprel_tm
from .wavepolarize_means import wavepolarize_means
from .psd import *
from .c_4_k import c_4_k
from .c_4_grad import c_4_grad
from .c_4_j import c_4_j
from .c_4_v import c_4_v
from .avg_4sc import avg_4sc
from .ebsp import ebsp
from .date_str import date_str

# 2020/09/04
from .find_closest import find_closest
from .cart2sph import cart2sph
from .corr_deriv import corr_deriv
from .eb_nrf import eb_nrf
from .find_closest import find_closest
from .lowpass import lowpass
from .mean import mean
from .poynting_flux import poynting_flux
from .remove_repeated_points import remove_repeated_points
from .solid_angle import solid_angle
from .wave_fft import wave_fft

from .pvi import pvi
from .pvi_4sc import pvi_4sc
from .pid_4sc import pid_4sc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pvi_4sc.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .norm import norm
from .dot import dot


def pvi_4sc(b_mms):
    """Compute the Partial Variance of Increments (PVI) using the definition in [12]_ as

    .. math::

            PVI_{ij}(t) = \\sqrt{\\frac{|\\Delta \\mathbf{B}_{ij}(t)|^2}{\\langle|\\Delta
            \\mathbf{B}_{ij}|^2\\rangle}}

    where :math:`\\Delta \\mathbf{B}_{ij}(t) = \\mathbf{B}_i(t) - \\mathbf{B}_i(t)` is the
    magnetic field increments, the average :math:`\\langle . \\rangle` is taken over the whole
    interval, and :math:`i`, :math:`j` = 1,2,3,4 is the MMS spacecraft number.

    In addition, computes, the rotation of the magnetic field between two spacecraft, i.e.,
    magnetic field shear angle, as :

    .. math::

        \\alpha_{ij}(t) = cos^{-1} \\frac{\\mathbf{B}_i(t).\\mathbf{B}_j(t)}
        {|\\mathbf{B}_i(t)| |\\mathbf{B}_j(t)|}


    Parameters
    ----------
    b_mms : list of xarray.DataArray
        List of the time series of the background magnetic field for the 4 spacecraft.

    Returns
    -------
    pvi_ij : xarray.DataArray
        Time series of the Partial Variance of Increments for the 6 pairs of spacecraft.

    alpha_ij : xarray.DataArray
        Time series of the magnetic field shear angle for the 6 pairs of spacecraft.

    References
    ----------
    .. [12] Chasapis, A., Retinó, A., Sahraoui, F., Vaivads, A., Khotyaintsev, Yu. V.,
            Sundkvist, D., et al. (2015) Thin current sheets and associated electron heating in
            turbulent space plasma. Astrophys. J. Lett. 804:L1.
            doi: https://doi.org/10.1088/2041-8205/804/1/L1


    """

    assert isinstance(b_mms, list) and isinstance(b_mms[0], xr.DataArray) and len(b_mms) == 4

    b_mms = [resample(b_xyz, b_mms[0]) for b_xyz in b_mms]

    i_indices = [0, 0, 0, 1, 1, 2]
    j_indices = [1, 2, 3, 2, 3, 3]

    delta_b_ij = [b_mms[i] - b_mms[j] for i, j in zip(i_indices, j_indices)]

    # Compute normalized partial variance of increments
    def pvi(b_i, b_j):
        return np.sqrt(norm(b_i - b_j) ** 2 / np.mean(norm(b_i - b_j) ** 2))

    pvi_ij = [pvi(b_mms[i], b_mms[j]) for i, j in zip(i_indices, j_indices)]

    # Compute magnetic field shear angle
    def shear_angle(b_i, b_j):
        return np.arccos(dot(b_i, b_j) / (norm(b_i) * norm(b_j)))

    theta_ij = [shear_angle(b_mms[i], b_mms[j]) for i, j in zip(i_indices, j_indices)]

    return pvi_ij, theta_ij



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
feeps_remove_sun.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy.time import Time

from .db_get_ts import db_get_ts
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv


def feeps_remove_sun(inp_dataset=None):
    """Removes the sunlight contamination from FEEPS data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset of energy spectrum of all eyes.

    Returns
    -------
    out : xarray.Dataset
        Dataset of cleaned energy spectrum of all eyes.

    See also
    --------
    pyrfu.mms.get_feeps_alleyes : Read energy spectrum for all FEEPS eyes.

    Examples
    --------
    >>> from pyrfu import mms
    >>> # Define time interval
    >>> tint = ["2017-07-18T13:04:00.000", "2017-07-18T13:07:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 2
    >>> # Load data from FEEPS
    >>> cps_i = mms.get_feeps_alleyes("CPSi_brst_l2", tint, mms_id)
    >>> cps_i_clean = mms.feeps_split_integral_ch(cps_i)
    >>> cps_i_clean_sun_removed = mms.feeps_remove_sun(cps_i_clean)

    """

    assert inp_dataset is not None and isinstance(inp_dataset, xr.Dataset)

    var = inp_dataset.attrs

    trange = list(Time(inp_dataset.time.data[[0, -1]], format="datetime64").isot)

    dset_name = "mms{:d}_feeps_{}_{}_{}".format(var["mmsId"], var["tmmode"], var["lev"],
                                                var["dtype"])
    dset_pref = "mms{:d}_epd_feeps_{}_{}_{}".format(var["mmsId"], var["tmmode"], var["lev"],
                                                    var["dtype"])

    spin_sectors = db_get_ts(dset_name, "_".join([dset_pref, "spinsectnum"]), trange)
    mask_sectors = read_feeps_sector_masks_csv(trange)

    outdict = {}

    for k in inp_dataset:
        outdict[k] = inp_dataset[k]
        if mask_sectors.get("mms{:d}_imask_{}".format(var["mmsId"], k)) is not None:
            bad_sectors = mask_sectors["mms{:d}_imask_{}".format(var["mmsId"], k)]

            for bad_sector in bad_sectors:
                this_bad_sector = np.where(spin_sectors == bad_sector)[0]
                if len(this_bad_sector) != 0:
                    outdict[k].data[this_bad_sector] = np.nan

    out = xr.Dataset(outdict, attrs=var)

    out.attrs = var

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
st_diff.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from pyrfu.pyrf import c_4_grad, gradient, avg_4sc, ts_vec_xyz


def st_diff(r_mms=None, b_mms=None, lmn=None):
    """Computes velocity of the structure using spatio-temporal derivative method [13]_ [14]_ as

    .. math::
        \\mathbf{V}_{str}^{\\mathbf{LMN}} = -\\textrm{d}_t \\mathbf{B}^{\\mathbf{LMN}}
        \\left [\\nabla^{\\mathbf{LMN}} \\mathbf{B}^{\\mathbf{LMN}}\\right]^T \\left [\\mathbf{
        S}^{\\mathbf{LMN}}\\right ]^{-1}

    where :math:`\\mathbf{B}^{\\mathbf{LMN}}`,
    :math:`\\nabla^{\\mathbf{LMN}}\\mathbf{B}^{\\mathbf{LMN}}`,
    :math:`\\mathbf{S}^{\\mathbf{LMN}}` and :math:`\\mathbf{V}_{str}^{\\mathbf{LMN}}` are namely
    the magnetic field, its gradient, its rotation rate tensor and the velocity of the structure
    in the LMN coordinates system.

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Spacecraft positions.

    b_mms : list of xarray.DataArray
        Background magnetic field.

    lmn : numpy.ndarray
        Structure coordinates system.

    Returns
    -------
    v_str : xarray.DataArray
        Velocity of the structure in its coordinates system.

    References
    ----------
    .. [13] Shi, Q. Q., Shen, C., Pu, Z. Y., Dunlop, M. W., Zong, Q. G., Zhang, H.,
            et al. (2005), Dimensional analysis of observed structures using multipoint magnetic
            field measurements: Application to Cluster. Geophysical Research Letters, 32, L12105.
            doi : https://doi.org/10.1029/2005GL022454.

    .. [14] Shi, Q. Q., Shen, C., Dunlop, M. W., Pu, Z. Y., Zong, Q. G., Liu, Z. X.,
            et al. (2006), Motion of observed structures calculated from multi‐point magnetic
            field measurements: Application to Cluster. Geophysical Research Letters, 33, L08109.
            doi : https://doi.org/10.1029/2005GL025073.
    """

    assert r_mms is not None and isinstance(r_mms, list)
    assert len(r_mms) == 4 and isinstance(r_mms[0], xr.DataArray)
    assert b_mms is not None and isinstance(b_mms, list)
    assert len(b_mms) == 4 and isinstance(b_mms[0], xr.DataArray)
    assert lmn is not None and isinstance(lmn, np.ndarray) and lmn.shape == (3, 3)

    # Compute magnetic field at the center of mass of the tetrahedron
    b_xyz = avg_4sc(b_mms)

    # Gradient of the magnetic field
    grad_b = c_4_grad(r_mms, b_mms)

    # Time derivative of the magnetic field at the center of mass of the tetrahedron
    db_dt = gradient(b_xyz)

    # Transform gradient to LMN frame
    l_grad_b = np.matmul(grad_b.data, lmn[:, 0])
    m_grad_b = np.matmul(grad_b.data, lmn[:, 1])
    n_grad_b = np.matmul(grad_b.data, lmn[:, 2])

    # Compute velocity of the structure using MDD
    v_str = np.zeros(db_dt.shape)
    v_str[:, 0] = np.sum(db_dt * l_grad_b, axis=1) / np.linalg.norm(l_grad_b, axis=1) ** 2
    v_str[:, 1] = np.sum(db_dt * m_grad_b, axis=1) / np.linalg.norm(m_grad_b, axis=1) ** 2
    v_str[:, 2] = np.sum(db_dt * n_grad_b, axis=1) / np.linalg.norm(n_grad_b, axis=1) ** 2

    # To time series
    v_str_xyz = ts_vec_xyz(b_xyz.time.data, -v_str)

    return v_str_xyz



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
remove_idist_background.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf.ts_tensor_xyz import ts_tensor_xyz


def remove_idist_background(n_i=None, v_gse_i=None, p_gse_i=None, n_bg_i=None, p_bg_i=None):
    """Removes penetrating radiation background from ion moments.

    Parameters
    ----------
    n_i : xarray.DataArray
        Time series of the ion density.

    v_gse_i : xarray.DataArray
        Time series of the ion bulk velocity.

    p_gse_i : xarray.DataArray
        Time series of the ion pressure tensor.

    n_bg_i : xarray.DataArray
        Time series of the background ion number density.

    p_bg_i : xarray.DataArray
        Time series of the background ion pressure scalar.

    Returns
    -------
    n_i_new : xarray.DataArray
        Time series of the corrected ion number density.

    v_gse_i_new : xarray.DataArray
        Time series of the corrected ion bulk velocity.

    p_gse_i : xarray.DataArray
        Time series of the corrected ion pressure tensor.

    References
    ----------
    MMS DIS Penetrating radiation correction methods :
    https://lasp.colorado.edu/galaxy/display/MFDPG/Penetrating+Radiation+in+DIS+Data

    """

    assert n_i is not None and isinstance(n_i, xr.DataArray)
    assert v_gse_i is not None and isinstance(v_gse_i, xr.DataArray)
    assert p_gse_i is not None and isinstance(p_gse_i, xr.DataArray)
    assert n_bg_i is not None and isinstance(n_bg_i, xr.DataArray)
    assert p_bg_i is not None and isinstance(p_bg_i, xr.DataArray)

    mp = constants.m_p.value

    # Number density
    n_i_new = n_i - n_bg_i.data

    # Bulk velocity
    v_gse_i_new = v_gse_i
    v_gse_i_new *= n_i / n_i_new

    # Pressure tensor
    p_gse_i_new_data = np.zeros(p_gse_i.shape)

    # P_xx_i
    p_gse_i_new_data[:, 0, 0] += p_gse_i.data[:, 0, 0]
    p_gse_i_new_data[:, 0, 0] -= p_bg_i.data
    p_gse_i_new_data[:, 0, 0] += mp * n_i.data * v_gse_i.data[:, 0] * v_gse_i.data[:, 0]
    p_gse_i_new_data[:, 0, 0] -= mp * n_i_new.data * v_gse_i_new.data[:, 0] * v_gse_i_new.data[:, 0]

    # P_yy_i
    p_gse_i_new_data[:, 1, 1] += p_gse_i.data[:, 1, 1]
    p_gse_i_new_data[:, 1, 1] -= p_bg_i.data
    p_gse_i_new_data[:, 1, 1] += mp * n_i.data * v_gse_i.data[:, 1] * v_gse_i.data[:, 1]
    p_gse_i_new_data[:, 1, 1] -= mp * n_i_new.data * v_gse_i_new.data[:, 1] * v_gse_i_new.data[:, 1]

    # P_zz_i
    p_gse_i_new_data[:, 2, 2] += p_gse_i.data[:, 2, 2]
    p_gse_i_new_data[:, 2, 2] -= p_bg_i.data
    p_gse_i_new_data[:, 2, 2] += mp * n_i.data * v_gse_i.data[:, 2] * v_gse_i.data[:, 2]
    p_gse_i_new_data[:, 2, 2] -= mp * n_i_new.data * v_gse_i_new.data[:, 2] * v_gse_i_new.data[:, 2]

    # P_xy_i & P_yx_i
    p_gse_i_new_data[:, 0, 1] += p_gse_i.data[:, 0, 1]
    p_gse_i_new_data[:, 0, 1] += mp * n_i.data * v_gse_i.data[:, 0] * v_gse_i.data[:, 1]
    p_gse_i_new_data[:, 0, 1] -= mp * n_i_new.data * v_gse_i_new.data[:, 0] * v_gse_i_new.data[:, 1]
    p_gse_i_new_data[:, 1, 0] = p_gse_i_new_data[:, 0, 1]

    # P_xz_i & P_zx_i
    p_gse_i_new_data[:, 0, 2] += p_gse_i.data[:, 0, 2]
    p_gse_i_new_data[:, 0, 2] += mp * n_i.data*v_gse_i.data[:, 0] * v_gse_i.data[:, 2]
    p_gse_i_new_data[:, 0, 2] -= mp * n_i_new.data * v_gse_i_new.data[:, 0] * v_gse_i_new.data[:, 2]
    p_gse_i_new_data[:, 2, 0] = p_gse_i_new_data[:, 0, 2]

    # P_yz_i & P_zy_i
    p_gse_i_new_data[:, 1, 2] += p_gse_i.data[:, 1, 2]
    p_gse_i_new_data[:, 1, 2] += mp * n_i.data * v_gse_i.data[:, 1] * v_gse_i.data[:, 2]
    p_gse_i_new_data[:, 1, 2] -= mp * n_i_new.data * v_gse_i_new.data[:, 1] * v_gse_i_new.data[:, 2]
    p_gse_i_new_data[:, 2, 1] = p_gse_i_new_data[:, 1, 2]

    p_gse_i_new = ts_tensor_xyz(p_gse_i.time.data, p_gse_i_new_data)

    return n_i_new, v_gse_i_new, p_gse_i_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
plot_line.py

@author : Louis RICHARD
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from cycler import cycler
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()

locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
date_form = mdates.ConciseDateFormatter(locator)
plt.style.use("seaborn-whitegrid")
sns.set_context("paper")

color = ["k", "tab:blue", "tab:red", "tab:green"]
# color = np.array([[0, 0, 0], [213, 94, 0], [0, 158, 115], [86, 180, 233]]) / 255

default_cycler = cycler(color=color)
plt.rc('axes', prop_cycle=default_cycler)


def plot_line(ax=None, inp=None, c=""):
    """Line plot of time series.

    Parameters
    ----------
    ax : to fill
        Axis

    inp : xarray.DataArray
        Time series to plot

    c : str
        Options

    """
    if ax is None:
        fig, ax = plt.subplots(1)
    if len(inp.shape) == 3:
        data = np.reshape(inp.data, (inp.shape[0], inp.shape[1] * inp.shape[2]))
    else:
        data = inp.data

    time = inp.time
    ax.plot(time, data, c)
    date_form = mdates.DateFormatter("%H:%M:%S")
    ax.xaxis.set_major_formatter(date_form)
    ax.grid(which="major", linestyle="-", linewidth="0.5", c="0.5")
    # ax.grid(which="minor",linestyle="-",linewidth="0.25")

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_feeps_omni.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr

from .get_feeps_active_eyes import get_feeps_active_eyes
from .get_feeps_oneeye import get_feeps_oneeye


# noinspection PyUnboundLocalVariable
def get_feeps_omni(tar_var="flux_ion_brst_l2", mms_id=1, tint=None, verbose=True):
    """Computes omni directional energy spectrum of the target data unit for the target specie
    over the target energy range.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like {data_unit}_{specie}_{data_rate}_{data_lvl}.

    tint : list of str
        Time interval.

    mms_id : int or float or str
        Index of the spacecraft.

    verbose : bool
        Set to True to follow the loading. Default is True.

    Returns
    --------
    out : xarray.DataArray
        Energy spectrum of the target data unit for the target specie in omni direction.

    """

    assert isinstance(tar_var, str)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)
    assert isinstance(verbose, bool)

    mms_id = int(mms_id)

    data_units = tar_var.split("_")[0]

    var = {"dtype": tar_var.split("_")[1], "tmmode": tar_var.split("_")[2],
           "lev": tar_var.split("_")[3]}

    specie = var["dtype"][0]

    if var["dtype"] == "electron":
        energies = np.array([33.2, 51.90, 70.6, 89.4, 107.1, 125.2, 146.5, 171.3,
                    200.2, 234.0, 273.4, 319.4, 373.2, 436.0, 509.2, 575.8])
    else:
        energies = np.array([57.9, 76.8, 95.4, 114.1, 133.0, 153.7, 177.6,
                    205.1, 236.7, 273.2, 315.4, 363.8, 419.7, 484.2,  558.6, 609.9])

    # set unique energy bins per spacecraft; from DLT on 31 Jan 2017
    e_corr = {"e": [14.0, -1.0, -3.0, -3.0], "i": [0.0, 0.0, 0.0, 0.0]}

    g_fact = {"e": [1.0, 1.0, 1.0, 1.0], "i": [0.84, 1.0, 1.0, 1.0]}

    energies += e_corr[specie][mms_id - 1]

    active_eyes = get_feeps_active_eyes(var, tint, mms_id)

    # percent error around energy bin center to accept data for averaging; 
    # anything outside of energies[i] +/- en_chk*energies[i] will be changed 
    # to NAN and not averaged   
    # en_chk = 0.1

    top_sensors = active_eyes["top"]
    bot_sensors = active_eyes["bottom"]

    top_it, bot_it = [{}, {}]

    for tsen in top_sensors:
        top = get_feeps_oneeye(f"{data_units}{var['dtype'][0]}_{var['tmmode']}_{var['lev']}",
                               mms_id, f"top-{tsen:d}", tint)

        mask = get_feeps_oneeye(f"mask{var['dtype'][0]}_{var['tmmode']}_{var['lev']}", mms_id,
                                f"top-{tsen:d}", tint)

        mask.data = np.tile(mask.data[:, 0], (mask.shape[1], 1)).T

        top.data[mask.data == 1] = np.nan

        top_it[tsen] = top

    for bsen in bot_sensors:
        bot = get_feeps_oneeye(f"{data_units}{var['dtype'][0]}_{var['tmmode']}_{var['lev']}",
                               mms_id, f"bottom-{bsen:d}", tint, verbose)

        mask = get_feeps_oneeye(f"mask{var['dtype'][0]}_{var['tmmode']}_{var['lev']}", mms_id,
                                f"bottom-{bsen:d}", tint, verbose)

        mask.data = np.tile(mask.data[:, 0], (mask.shape[1], 1)).T

        bot.data[mask.data == 1] = np.nan

        bot_it[bsen] = bot

    dalleyes = np.empty((top.shape[0], top.shape[1], len(top_sensors) + len(bot_sensors)))

    for i, tsen in enumerate(top_sensors):
        dalleyes[..., i] = top_it[tsen]

    for i, bsen in enumerate(bot_sensors):
        dalleyes[..., len(top_sensors) + i] = bot_it[tsen]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        flux_omni = np.nanmean(dalleyes, axis=2)

    flux_omni *= g_fact[specie][mms_id - 1]

    time = top_it[top_sensors[0]]

    out = xr.DataArray(flux_omni[:], coords=[time, energies], dims=["time", "energy"])

    return out



import pathlib
from setuptools import setup, find_packages
from sphinx.setup_command import BuildDoc

HERE = pathlib.Path(__file__).parent

VERSION = "1.8.8"
PACKAGE_NAME = "pyrfu"
AUTHOR = "Louis RICHARD"
AUTHOR_EMAIL = "louir@irfu.se"
URL = "https://github.com/louis-richard/irfu-python"

LICENSE = "MIT License"
DESCRIPTION = "Python Space Physics Environment Data Analysis"

with open("README.rst", "r") as fh:
    LONG_DESCRIPTION = fh.read()


INSTALL_REQUIRES = [
      "astropy",
      "cycler",
      "python-dateutil",
      "matplotlib",
      "numpy",
      "sphinx>=1.4",
      "pandas",
      "pvlib",
      "pyfftw",
      "scipy",
      "seaborn",
      "sfs",
      "tqdm",
      "xarray",
      "sphinx_rtd_theme", 
      "ipython", 
      "ipykernel", 
      "nbsphinx",
      "spacepy",
      "sphinxcontrib-apidoc"]

PYTHON_REQUIRES = '>=3.7'


cmdclass = {'build_sphinx': BuildDoc}

setup(
      name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      python_requires=PYTHON_REQUIRES,
      packages=find_packages(),
      include_package_data=True,
      cmdclass=cmdclass,
      # these are optional and override conf.py settings
      command_options={
        'build_sphinx': {
            'project': ('setup.py', PACKAGE_NAME),
            'version': ('setup.py', VERSION),
            'release': ('setup.py', VERSION),
            'source_dir': ('setup.py', 'docs')}},
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
cross.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .ts_vec_xyz import ts_vec_xyz


def cross(x=None, y=None):
    """Computes cross product of two fields.

    Parameters
    ----------
    x : xarray.DataArray
        Time series of the first field X.

    y : xarray.DataArray
        Time series of the second field Y.

    Returns
    -------
    out : xarray.DataArray
        Time series of the cross product Z = XxY.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Define time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Index of the MMS spacecraft
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)
    >>> # Compute magnitude of the magnetic field
    >>> b_mag = pyrf.norm(b_xyz)
    >>> # Compute ExB drift velocity
    >>> v_xyz_exb = pyrf.cross(e_xyz, b_xyz) / b_mag ** 2

    """

    assert x is not None and isinstance(x, xr.DataArray)
    assert y is not None and isinstance(y, xr.DataArray)

    if len(x) != len(y):
        y = resample(y, x)

    out_data = np.cross(x.data, y.data, axis=1)

    out = ts_vec_xyz(x.time.data, out_data)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_scalar.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def ts_scalar(t=None, data=None, attrs=None):
    """Create a time series containing a 0th order tensor

    Parameters
    ----------
    t : numpy.ndarray
        Array of times.

    data : numpy.ndarray
        Data corresponding to the time list.

    attrs : dict
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        0th order tensor time series.

    """

    assert t is not None and isinstance(t, np.ndarray)
    assert data is not None and isinstance(data, np.ndarray)

    if data.ndim != 1:
        raise TypeError("Input must be a scalar")

    if len(t) != len(data):
        raise IndexError("Time and data must have the same length")

    flag_attrs = True

    if attrs is None:
        flag_attrs = False

    out = xr.DataArray(data, coords=[t[:]], dims="time")

    if flag_attrs:
        out.attrs = attrs

    out.attrs["TENSOR_ORDER"] = 0

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
normalize.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np


def normalize(inp=None):
    """Normalizes the input field.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field.

    Returns
    -------
    out : xarray.DataArray
        Time series of the normalized input field.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> # Compute the normalized magnetic field
    >>> b = pyrf.normalize(b_xyz)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    out = inp / np.linalg.norm(inp, axis=1, keepdims=True)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
resample.py

@author : Louis RICHARD
"""

import bisect
import warnings
import numpy as np

from scipy import interpolate
import xarray as xr


def resample(inp=None, ref=None, **kwargs):
    """
    Resample inp to the time line of ref. If sampling of X is more than two times higher than Y,
    we average X, otherwise we interpolate X.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to resample.

    ref : xarray.DataArray
        Reference time line.

    **kwargs : dict
        Hash table of keyword arguments with :
            * method : str
                Method of interpolation "spline", "linear" etc. (default "linear") if method is
                given then interpolate independent of sampling.

            * fs : float
                Sampling frequency of the Y signal, 1/window.

            * window : int or float or np.ndarray
                Length of the averaging window, 1/fsample.

            * fs : str
                Sampling frequency of the Y signal, 1/window.

            * mean : bool
                Use mean when averaging.

            * median : bool
                Use median instead of mean when averaging.

            * max : bool
                Use max instead of mean when averaging.

    Returns
    -------
    out : xarray.DataArray
        Resampled input to the reference time line using the selected method.


    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)
    >>> # Resample magnetic field to electric field sampling
    >>> b_xyz = pyrf.resample(b_xyz, e_xyz)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert ref is not None and isinstance(ref, xr.DataArray)

    sfy, thresh, method = [[], 0, ""]

    flag_do = "check"

    median_flag, mean_flag, max_flag = [False, False, False]

    if "method" in kwargs:
        if isinstance(kwargs["method"], str):
            method = kwargs["method"]
        else:
            raise TypeError("METHOD must be string")

    if "fs" in kwargs:
        if sfy:
            raise ValueError("fs/window already specified")

        if isinstance(kwargs["fs"], (float, int)):
            sfy = kwargs["fs"]
        else:
            raise TypeError("fs must be numeric")

    if "window" in kwargs:
        if sfy:
            raise ValueError("fs/window already specified")

        if (not (not isinstance(kwargs["window"], int) and not isinstance(kwargs["window"], float)
                 and not isinstance(kwargs["window"], np.ndarray))):
            sfy = 1 / kwargs["window"]
        else:
            raise TypeError("METHOD must be numeric")

    if "mean" in kwargs:
        if kwargs["mean"]:
            mean_flag = True

            flag_do = "average"

    if "median" in kwargs:
        if kwargs["median"]:
            median_flag = True

            flag_do = "average"

    if "max" in kwargs:
        if kwargs["max"]:
            max_flag = True

            flag_do = "average"

    inp_time = inp.time.data.view("i8") * 1e-9
    ref_time = ref.time.data.view("i8") * 1e-9
    inp_data = inp.data

    if len(inp) == 1:
        if len(inp.shape) == 1:
            out_data = np.tile(inp_data, len(ref_time))
        else:
            out_data = np.tile(inp_data, (len(ref_time), 1))

    n_data = len(ref_time)

    if flag_do == "check":
        if n_data > 1:
            if not sfy:
                sfy1 = 1 / (ref_time[1] - ref_time[0])

                if n_data == 2:
                    sfy = sfy1

                    not_found = False
                else:
                    not_found = True

                cur, max_try = [2, 10]

                while not_found and cur <= n_data and cur - 3 < max_try:
                    sfy = 1 / (ref_time[cur] - ref_time[cur - 1])

                    if np.absolute(sfy-sfy1) < sfy*0.001:
                        not_found = False

                        sfy = (sfy + sfy1) / 2

                        break

                    sfy = sfy1

                    cur += 1

                if not_found:
                    raise RuntimeError(
                        "Cannot guess sampling frequency. Tried {:d} times".format(max_try))

                del sfy1

            if len(inp_time) / (inp_time[-1] - inp_time[0]) > 2 * sfy:
                flag_do = "average"
                warnings.warn("Using averages in resample", UserWarning)
            else:
                flag_do = "interpolation"
        else:
            flag_do = "interpolation"  # If one output time then do interpolation

    if flag_do == "average":
        if method:
            raise ValueError("cannot mix interpolation and averaging flags")

        if not sfy:
            sfy1 = 1 / (ref_time[1] - ref_time[0])

            if n_data == 2:
                sfy = sfy1

                not_found = False
            else:
                not_found = True

            cur, max_try = [2, 10]

            while not_found and cur <= n_data and cur - 3 < max_try:
                sfy = 1 / (ref_time[cur] - ref_time[cur - 1])

                if np.absolute(sfy - sfy1) < sfy * .001:
                    not_found = False

                    sfy = (sfy + sfy1) / 2
                    break

                sfy = sfy1
                cur += 1

            if not_found:
                raise RuntimeError(
                    "Cannot guess sampling frequency. Tried {:d} times".format(max_try))

            del sfy1

        dt2 = .5 / sfy  # Half interval

        inp_shape = list(inp_data.shape)
        inp_shape[0] = n_data

        inp_shape = tuple(inp_shape)

        out_data = np.zeros(inp_shape)

        for i in range(n_data):
            idx_l = bisect.bisect_left(inp_time, ref_time[i] - dt2)
            idx_r = bisect.bisect_right(inp_time, ref_time[i] + dt2)

            ii = np.arange(idx_l, idx_r)

            if ii.size == 0:
                out_data[i, ...] = np.nan
            else:
                if thresh:
                    stddev = np.std(inp_data[ii, ...], axis=0)

                    mm = np.mean(inp_data[ii, ...], axis=0)

                    if any(np.isnan(stddev)):
                        for k, stdd in enumerate(stddev):
                            if not np.isnan(stdd):
                                kk = bisect.bisect_right(inp_data[ii, k + 1] - mm[k], thresh * stdd)
                                if kk:
                                    if median_flag:
                                        out_data[i, k + 1] = np.median(inp_data[ii[kk], k + 1],
                                                                       axis=0)
                                    elif max_flag:
                                        out_data[i, k + 1] = np.max(inp_data[ii[kk], k + 1], axis=0)
                                    else:
                                        out_data[i, k + 1] = np.mean(inp_data[ii[kk], k + 1],
                                                                     axis=0)
                                else:
                                    out_data[i, k + 1] = np.nan
                            else:
                                out_data[i, ...] = np.nan
                else:
                    if median_flag:
                        out_data[i, ...] = np.median(inp_data[ii, ...], axis=0)
                    elif max_flag:
                        out_data[i, ...] = np.max(inp_data[ii, ...], axis=0)
                    else:
                        out_data[i, ...] = np.mean(inp_data[ii, ...], axis=0)

    elif flag_do == "interpolation":
        if any([mean_flag, median_flag, max_flag]):
            raise ValueError("cannot mix interpolation and averaging flags")

        if not method:
            method = "linear"

        # If time series agree, no interpolation is necessary.
        if len(inp_time) == len(ref_time) and all(inp_time == ref_time):
            out_data = inp_data
            coord = [ref.coords["time"].data]

            if len(inp.coords) > 1:
                for k in inp.dims[1:]:
                    coord.append(inp.coords[k].data)

            out = xr.DataArray(out_data, coords=coord, dims=inp.dims, attrs=inp.attrs)

            return out

        tck = interpolate.interp1d(inp_time, inp_data, kind=method, axis=0,
                                   fill_value="extrapolate")
        out_data = tck(ref_time)

    else:
        raise NameError("Invalid method")

    coord = [ref.coords["time"]]

    if len(inp.coords) > 1:
        for k in inp.dims[1:]:
            coord.append(inp.coords[k].data)

    out = xr.DataArray(out_data, coords=coord, dims=inp.dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
c_4_v.py

@author : Louis RICHARD
"""

import numpy as np

from scipy import interpolate
from astropy.time import Time


def get_vol_ten(r=None, t=None):
    """to fill."""

    if len(t) == 1:
        t = np.array([t, t, t, t])

    tckr_x, tckr_y, tckr_z = [[], [], []]

    for i in range(4):
        tckr_x.append(interpolate.interp1d(r[i].time.data, r[i].data[:, 0]))
        tckr_y.append(interpolate.interp1d(r[i].time.data, r[i].data[:, 1]))
        tckr_z.append(interpolate.interp1d(r[i].time.data, r[i].data[:, 2]))

        r[i] = np.array([tckr_x[i](t[0]), tckr_y[i](t[0]), tckr_z[i](t[0])])

    # Volumetric tensor with SC1 as center.
    dr_mat = (np.vstack(r[1:]) - np.tile(r[0], (3, 1))).T

    return dr_mat


def c_4_v(r=None, x=None):
    """Calculates velocity or time shift of discontinuity as in [6]_.

    Parameters
    ----------
    r : list of xarray.DataArray
        Time series of the positions of the spacecraft.

    x : list
        Crossing times or time and velocity.

    Returns
    -------
    out : numpy.ndarray
        Discontinuity velocity or time shift with respect to mms1.

    References
    ----------
    .. [6]	Vogt, J., Haaland, S., and Paschmann, G. (2011) Accuracy of multi-point boundary
            crossing time analysis, Ann. Geophys., 29, 2239-2252,
            doi : https://doi.org/10.5194/angeo-29-2239-2011

    """
    if isinstance(x, np.ndarray) and x.dtype == np.datetime64:
        flag = "v_from_t"

        x = Time(x, format="datetime64").unix
    elif x[1] > 299792.458:
        flag = "v_from_t"
    else:
        flag = "dt_from_v"

    if flag.lower() == "v_from_t":
        # Time input, velocity output
        t = x
        dr_mat = get_vol_ten(r, t)
        tau = np.array(t[1:]) - t[0]
        m = np.linalg.solve(dr_mat, tau)

        # "1/v vector"
        out = m / np.linalg.norm(m) ** 2

    elif flag.lower() == "dt_from_v":
        # Time and velocity input, time output
        tc = x[0]  # center time
        v = np.array(x[1:])  # Input velocity
        m = v / np.linalg.norm(v) ** 2

        dr_mat = get_vol_ten(r, tc)

        dt = np.matmul(dr_mat, m)
        out = np.hstack([0, dt])

    else:
        raise ValueError("Invalid flag")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_skymap.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def ts_skymap(t=None, data=None, energy=None, phi=None, theta=None, **kwargs):
    """Creates a skymap of the distribution function.

    Parameters
    ----------
    t : numpy.ndarray
        List of times.

    data : numpy.ndarray
        Values of the distribution function.

    energy : numpy.ndarray
        Energy levels.

    phi : numpy.ndarray
        Azimuthal angles.

    theta : numpy.ndarray
        Elevation angles.

    **kwargs : dict
        Hash table of keyword arguments with :
            * energy0 : numpy.ndarray
                Energy table 0 (odd time indices).

            * energy1 : numpy.ndarray
                Energy table 1 (even time indices).

            * esteptable : numpy.ndarray
                Time series of the stepping table between energies (burst).


    Returns
    -------
    out : xarray.DataArray
        Skymap of the distribution function.

    """

    assert t is not None and isinstance(t, np.ndarray)
    assert data is not None and isinstance(data, xr.DataArray)
    assert phi is not None and isinstance(phi, np.ndarray)
    assert theta is not None and isinstance(theta, np.ndarray)

    if energy is None:
        if "energy0" in kwargs:
            energy0, energy0_ok = [kwargs["energy0"], True]
        else:
            energy0, energy0_ok = [None, False]

        if "energy1" in kwargs:
            energy1, energy1_ok = [kwargs["energy1"], True]
        else:
            energy1, energy1_ok = [None, False]

        if "esteptable" in kwargs:
            esteptable, esteptable_ok = [kwargs["esteptable"], True]
        else:
            esteptable, esteptable_ok = [None, False]

        if not energy0_ok and not energy1_ok and not esteptable_ok:
            raise ValueError("Energy input required")

        energy = np.tile(energy0, (len(esteptable), 1))

        energy[esteptable == 1] = np.tile(energy1, (int(np.sum(esteptable)), 1))

    else:
        energy0, energy1, esteptable = [None] * 3

        energy0_ok, energy1_ok, esteptable_ok = [False] * 3

    if phi.ndim == 1:
        phi = np.tile(phi, (len(t), 1))

    out_dict = {"data": (["time", "idx0", "idx1", "idx2"], data), "phi": (["time", "idx1"], phi),
                "theta": (["idx2"], theta), "energy": (["time", "idx0"], energy), "time": t,
                "idx0": np.arange(32),
                "idx1": np.arange(32), "idx2": np.arange(16)}

    out = xr.Dataset(out_dict)

    if energy0_ok:
        out.attrs["energy0"] = energy0

    if energy1_ok:
        out.attrs["energy1"] = energy1

    if energy0_ok:
        out.attrs["esteptable"] = esteptable

    return out



# -*- coding: utf-8 -*-
"""
__init__.py

@author : Louis RICHARD
"""

from .split_vs import split_vs
from .list_files import list_files
from .get_ts import get_ts
from .get_dist import get_dist
from .get_data import get_data
from .db_get_ts import db_get_ts

# Wave analysis
from .fk_power_spectrum_4sc import fk_power_spectrum_4sc
from .lh_wave_analysis import lh_wave_analysis
from .whistler_b2e import whistler_b2e

# FEEPS
from .get_feeps_energy_table import get_feeps_energy_table
from .get_feeps_active_eyes import get_feeps_active_eyes
from .get_feeps_oneeye import get_feeps_oneeye
from .get_feeps_omni import get_feeps_omni
from .get_feeps_alleyes import get_feeps_alleyes
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv
from .feeps_split_integral_ch import feeps_split_integral_ch
from .feeps_remove_sun import feeps_remove_sun
from .calc_feeps_omni import calc_feeps_omni
from .feeps_spin_avg import feeps_spin_avg
from .feeps_pitch_angles import feeps_pitch_angles
from .calc_feeps_pad import calc_feeps_pad
from .get_eis_allt import get_eis_allt
from .get_eis_omni import get_eis_omni
from .remove_idist_background import remove_idist_background
from .psd_moments import psd_moments
from .rotate_tensor import rotate_tensor


# 2020-09-09
from .calculate_epsilon import calculate_epsilon
from .dft_time_shift import dft_time_shift
from .estimate_phase_speed import estimate_phase_speed
from .fft_bandpass import fft_bandpass
from .get_pitch_angle_dist import get_pitch_angle_dist
from .make_model_vdf import make_model_vdf
from .psd_rebin import psd_rebin

from .load_ancillary import load_ancillary
from .vdf_omni import vdf_omni
from .spectr_to_dataset import spectr_to_dataset
from .vdf_to_deflux import vdf_to_deflux
from .vdf_to_dpflux import vdf_to_dpflux
from .dpflux_to_vdf import dpflux_to_vdf
from .deflux_to_vdf import deflux_to_vdf

from .fk_power_spectrum import fk_power_spectrum
from .lh_wave_analysis import lh_wave_analysis
from .copy_files import copy_files


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mva.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np


def mva(inp=None, flag="mvar"):
    """Compute the minimum variance frame.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the quantity to find minimum variance frame.

    flag : {"mvar", "<bn>=0", "td"}
        Constrain.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input quantity in LMN coordinates.

    l : numpy.ndarray
        Eigenvalues l[0]>l[1]>l[2].

    lmn : numpy.ndarray
        Eigenvectors LMN coordinates.

    See also
    --------
    pyrfu.pyrf.new_xyz

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> # Compute MVA frame
    >>> b_lmn, lamb, frame = pyrf.mva(b_xyz)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert isinstance(flag, str)

    inp_data = inp.data

    inp_m = np.mean(inp_data, 0)

    if flag in ["mvar", "<bn>=0"]:
        m_mu_nu_m = np.mean(inp_data[:, [0, 1, 2, 0, 0, 1]] * inp_data[:, [0, 1, 2, 1, 2, 2]], 0)
        m_mu_nu_m -= inp_m[[0, 1, 2, 0, 0, 1]] * inp_m[[0, 1, 2, 1, 2, 2]]

    elif flag.lower() == "td":
        m_mu_nu_m = np.mean(inp_data[:, [0, 1, 2, 0, 0, 1]] * inp_data[:, [0, 1, 2, 1, 2, 2]], 0)

    else:
        raise ValueError("invalid flag")

    m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]])

    # Compute eigenvalues and eigenvectors
    [l, lmn] = np.linalg.eig(m_mu_nu)

    # Sort eigenvalues
    idx = l.argsort()[::-1]
    l, lmn = [l[idx], lmn[:, idx]]

    # ensure that the frame is right handed
    lmn[:, 2] = np.cross(lmn[:, 0], lmn[:, 1])

    if flag.lower() == "<bn>=0":
        inp_mvar_mean = np.mean(
            np.sum(
                np.tile(inp_data, (3, 1, 1)) * np.transpose(np.tile(lmn, (inp_data.shape[0], 1, 1)),
                                                            (2, 0, 1)), 1), 1)

        a = np.sum(inp_mvar_mean ** 2)

        b = -(l[1] + l[2]) * inp_mvar_mean[0] ** 2
        b -= (l[0] + l[2]) * inp_mvar_mean[1] ** 2
        b -= (l[0] + l[1]) * inp_mvar_mean[2] ** 2

        c = l[1] * l[2] * inp_mvar_mean[0] ** 2
        c += l[0] * l[2] * inp_mvar_mean[1] ** 2
        c += l[0] * l[1] * inp_mvar_mean[2] ** 2

        r = np.roots([a, b, c])

        l_min = np.min(r)

        n = inp_mvar_mean / (l - l_min)
        n /= np.linalg.norm(n, keepdims=True)
        n = np.matmul(lmn, n)

        bn = np.sum(inp_data * np.tile(n, (inp_data.shape[0], 1)), axis=1)

        inp_data_2 = inp_data - np.tile(bn, (3, 1)).T * np.tile(n, (inp_data.shape[0], 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, [0, 1, 2, 0, 0, 1]] * inp_data_2[:, [0, 1, 2, 1, 2, 2]],
                            0)
        m_mu_nu_m -= inp_data_2_m[[0, 1, 2, 0, 0, 1]] * inp_data_2_m[[0, 1, 2, 1, 2, 2]]

        m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]])

        l, lmn = np.linalg.eig(m_mu_nu)

        idx = l.argsort()[::-1]

        l, lmn = [l[idx], lmn[:, idx]]

        l[2], lmn[:, 2] = [l_min, np.cross(lmn[:, 0], lmn[:, 1])]

    elif flag.lower() == "td":
        ln = l[2]
        bn = np.sum(inp_data * np.tile(lmn[:, 2], (inp_data.shape[0], 1)), axis=1)

        inp_data_2 = inp_data - np.tile(bn, (3, 1)).T * np.tile(lmn[:, 2], (inp_data.shape[0], 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, [0, 1, 2, 0, 0, 1]] * inp_data_2[:, [0, 1, 2, 1, 2, 2]],
                            0)
        m_mu_nu_m -= inp_data_2_m[[0, 1, 2, 0, 0, 1]] * inp_data_2_m[[0, 1, 2, 1, 2, 2]]

        m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]])

        l, lmn = np.linalg.eig(m_mu_nu)

        idx = l.argsort()[::-1]

        l, lmn = [l[idx], lmn[:, idx]]

        l[2], lmn[:, 2] = [ln, np.cross(lmn[:, 0], lmn[:, 1])]

    out_data = (lmn.T @ inp_data.T).T

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)

    return out, l, lmn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
feeps_spin_avg.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr
from astropy.time import Time

from .db_get_ts import db_get_ts


def feeps_spin_avg(inp_dataset_omni=None):
    """This function will spin-average the omni-directional FEEPS energy spectra.
    
    Parameters
    ----------
    inp_dataset_omni : xarray.DataArray
        Spectrogram of all eyes in OMNI.

    Returns
    -------
    out : xarray.DataArray
        Spin-averaged OMNI energy spectrum.

    """

    if inp_dataset_omni is None:
        raise ValueError("feeps_spin_avg requires at least one argument")

    var = inp_dataset_omni.attrs

    # get the spin sectors
    # v5.5+ = mms1_epd_feeps_srvy_l1b_electron_spinsectnum
    trange = list(Time(inp_dataset_omni.time.data[[0, -1]], format="datetime64").isot)

    dset_name = f"mms{var['mmsId']}_feeps_{var['tmmode']}_{var['lev']}_{var['dtype']}"
    dset_pref = f"mms{var['mmsId']}_epd_feeps_{var['tmmode']}_{var['lev']}_{var['dtype']}"

    spin_sectors = db_get_ts(dset_name, "_".join([dset_pref, "spinsectnum"]), trange)

    spin_starts = [spin_end + 1 for spin_end in np.where(spin_sectors[:-1] >= spin_sectors[1:])[0]]

    energies = inp_dataset_omni.coords["energy"]
    data = inp_dataset_omni.data

    spin_avg_flux = np.zeros([len(spin_starts), len(energies)])

    current_start = spin_starts[0]

    for spin_idx in range(1, len(spin_starts) - 1):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            spin_avg_flux[spin_idx - 1, :] = np.nanmean(
                data[current_start:spin_starts[spin_idx] + 1, :], axis=0)
        
        current_start = spin_starts[spin_idx] + 1

    time = inp_dataset_omni.coords["time"].data[spin_starts]
    out = xr.DataArray(spin_avg_flux, coords=[time, energies], dims=inp_dataset_omni.dims)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_time.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy.time import Time


def ts_time(t=None, fmt="unix"):
    """Creates time line in DataArray.

    Parameters
    ----------
    t : numpy.ndarray
        Input time line.

    fmt : str
        Format of the input time line.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time line.

    """

    assert t is not None and isinstance(t, np.ndarray)

    t = Time(t, format=fmt).datetime64

    out = xr.DataArray(t, coords=[t], dims=["time"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
db_get_ts.py

@author : Louis RICHARD
"""

from .list_files import list_files
from .get_ts import get_ts
from ..pyrf import ts_append


# noinspection PyUnboundLocalVariable
def db_get_ts(dataset_name="", cdf_name="", tint=None):
    """Get variable time series in the cdf file.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset.

    cdf_name : str
        Name of the target field in cdf file.

    tint : list of str
        Time interval.

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable.

    """

    assert isinstance(dataset_name, str)
    assert isinstance(cdf_name, str)
    assert tint is not None and isinstance(tint, list)

    dataset = dataset_name.split("_")

    # Index of the MMS spacecraft
    probe = dataset[0][-1]

    var = {"inst": dataset[1], "tmmode": dataset[2], "lev": dataset[3]}

    try:
        var["dtype"] = dataset[4]
    except IndexError:
        pass

    files = list_files(tint, probe, var)

    for i, file in enumerate(files):
        temp = get_ts(file, cdf_name, tint)
        if i == 0:
            out = temp
        else:
            out = ts_append(out, temp)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
agyro_coeff.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def agyro_coeff(p_xyz=None):
    """Computes agyrotropy coefficient as in [1]_

    .. math::

        Q = \\frac{P_{12}^2 + P_{13}^2 + P_{23}^2}{P_\\perp^2 + 2 P_\\perp P_\\parallel}


    Parameters
    ----------
    p_xyz : xarray.DataArray
        Time series of the pressure tensor

    Returns
    -------
    sqrt_q : xarray.DataArray
        Time series of the agyrotropy coefficient of the specie

    References
    ----------
    .. [1]  Swisdak, M. (2016), Quantifying gyrotropy in magnetic reconnection, Geophys. Res.
            Lett., 43, 43–49, doi: https://doi.org/10.1002/2015GL066980.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> Tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> ic = 1
    >>> # Load magnetic field and electron pressure tensor
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2",Tint,1)
    >>> p_xyz_e = mms.get_data("Pe_gse_fpi_fast_l2",Tint,1)
    >>> # Rotate electron pressure tensor to field aligned coordinates
    >>> p_xyzfac_e = mms.rotate_tensor(p_xyz_e,"fac",b_xyz,"pp")
    >>> # Compute agyrotropy coefficient
    >>> sqrt_q_e = pyrf.agyro_coeff(p_xyzfac_e)

    """

    if p_xyz is None:
        raise ValueError("agyro_coeff requires at least one argument")

    if not isinstance(p_xyz, xr.DataArray):
        raise TypeError("Input must be a DataArray")

    if p_xyz.ndim != 3:
        raise TypeError("Input must be a second order tensor")

    # Parallel and perpendicular components
    p_para, p_perp = [p_xyz[:, 0, 0], (p_xyz[:, 1, 1] + p_xyz[:, 2, 2]) / 2]

    # Off-diagonal terms
    p_12, p_13, p_23 = [p_xyz[:, 0, 1], p_xyz[:, 0, 2], p_xyz[:, 1, 2]]

    sqrt_q = np.sqrt((p_12 ** 2 + p_13 ** 2 + p_23 ** 2) / (p_perp ** 2 + 2 * p_perp * p_para))

    return sqrt_q



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
e_vxb.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .ts_vec_xyz import ts_vec_xyz


def e_vxb(v=None, b=None, flag="vxb"):
    """Computes the convection electric field :math:`\\mathbf{V}\\times\\mathbf{B}` (default) or the
    :math:`\\mathbf{E}\\times\\mathbf{B}/|\\mathbf{B}|^{2}` drift velocity (flag="exb").

    Parameters
    ----------
    v : xarray.DataArray
        Time series of the velocity/electric field.

    b : xarray.DataArray
        Time series of the magnetic field.

    flag : str
        Method flag :
            * "vxb" : computes convection electric field (default).
            * "exb" : computes ExB drift velocity.

    Returns
    -------
    out : xarray.DataArray
        Time series of the convection electric field/ExB drift velocity.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)
    >>> # Compute ExB drift velocity
    >>> v_xyz_exb = pyrf.e_vxb(e_xyz, b_xyz,"ExB")

    """

    assert v is not None and isinstance(v, xr.DataArray)
    assert b is not None and isinstance(b, xr.DataArray)

    if flag.lower() == "exb":
        estimate_exb = True
        estimate_vxb = False
    else:
        estimate_exb = False
        estimate_vxb = True

    if v.ndim == 1 and len(v) == 3:
        input_v_cons = True
    else:
        input_v_cons = False

    if estimate_exb:
        e = v

        if len(e) != len(b):
            b = resample(b, e)

        res = 1e3 * np.cross(e.data, b.data, axis=1)
        res /= np.linalg.norm(b.data, axis=1)[:, None] ** 2

        attrs = {"UNITS": "km/s", "FIELDNAM": "Velocity", "LABLAXIS": "V"}

    elif estimate_vxb:
        if input_v_cons:
            res = np.cross(np.tile(v, (len(b), 1)), b.data) * (-1) * 1e-3

        else:
            if len(v) != len(b):
                b = resample(b, v)

            res = np.cross(v.data, b.data) * (-1) * 1e-3

        attrs = {"UNITS": "mV/s", "FIELDNAM": "Electric field", "LABLAXIS": "E"}

    else:
        raise ValueError("Invalid flag")

    out = ts_vec_xyz(b.time.data, res, attrs)

    return out
